import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:ui';

import 'package:advertising_id/advertising_id.dart';
import 'package:appsflyer_sdk/appsflyer_sdk.dart';
import 'package:device_info/device_info.dart';
import 'package:firebase_analytics/firebase_analytics.dart';
import 'package:firebase_core/firebase_core.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/services.dart';
import 'package:http/http.dart' as http;
import 'package:intl/intl.dart';
import 'package:lemnisk_plugin/AttributeBuilder.dart';
import 'package:package_info_plus/package_info_plus.dart';
import 'package:platform_device_id_platform_interface/platform_device_id_platform_interface.dart';
import 'package:uuid/uuid.dart';
import 'package:clevertap_plugin/clevertap_plugin.dart';
// import 'package:workmanager/workmanager.dart';

import 'LemniskFlutterCore.dart';
import 'firebase/LemniskFirebase.dart';
import 'model/GenricModel.dart';

class LemniskHelper {
  static FirebaseApp? _appFirebase;
  static AppsflyerSdk? _appsflyerSdk;

  static String? _writeKey;
  static String? _serverUrl;

  //firebase App Core
  static bool? _enableFirebase;
  static String? _projectName;
  static String? _appId;
  static String? _apiKey;
  static String? _projectId;
  static String? _messagingSenderId;

  static bool? _enableAppsFlyer;
  static String? _afDevKey;
  static String? _afAppId;
  static FirebaseAnalytics? analytics;

  static bool? _enableClevertap;
  static late CleverTapPlugin? _clevertapPlugin = new CleverTapPlugin();

  static final String _transformUrl = "https://mu-tr.lemnisk.co/v3/datasources/";
  static final writeKeytoConfigMapping = '{"ga4":{"signUp":"sign_up","signIn":"login","searchSubmit":{"search":{"search_term":"properties.searchTerm"}},"serviceListViewed":{"view_item_list":{"items":[{"item_category":"properties.serviceCategory"}]}},"serviceDetailViewed":{"view_item":{"items":[{"item_category":"properties.serviceCategory","item_category2":"properties.serviceSubcategory"}]}},"serviceClicked":{"select_item":{"items":[{"item_category":"properties.serviceCategory","item_category2":"properties.serviceSubcategory"}]}},"bookServiceClicked":{"begin_checkout":{"currency":"properties.currency","value":"properties.value","items":[{"item_name":"properties.serviceItem","item_category":"properties.serviceCategory","item_category2":"properties.serviceSubcategory"}]}},"bookingCompleted":{"purchase":{"currency":"properties.currency","value":"properties.value","items":[{"item_id":"properties.projectID","item_name":"properties.serviceItem","item_category":"properties.serviceCategory","item_category2":"properties.serviceSubcategory"}]}}},"appsflyer":{"event":{"signUp":"af_complete_registration","signIn":"af_login","searchSubmit":"af_search","serviceListViewed":"af_list_view","serviceDetailViewed":"af_content_view","bookServiceClicked":"af_initiated_checkout","bookingCompleted":"af_purchase"},"properties":{"revenue":"af_revenue","value":"af_price","currency":"af_currency","searchTerm":"af_search_string","serviceCategory":"af_content_type","serviceItem":"af_content_id","projectID":"af_receipt_id"}},"clevertap":{"chargeDetails":["currency","value","bookingPaymentType"],"chargeEvents":["bookingCompleted"]}}';
  static Map? parsedWriteKeytoConfigMap;// = jsonDecode(writeKeytoConfigMapping);
  static const simplePeriodic30minsTask = "simplePeriodic30minsTask";
  static final sdkVersion = "0.0.9";

  /*static void callbackDispatcher() {
    Workmanager().executeTask((task, inputData) async {
      LemniskHelper.getTransformConfig();
      return Future.value(true);
    });
  }*/

  static Future<String?> configure(String? serverUrl, String? writeKey,
      LemniskFirebaseCore? firebase, LemniskApsflyerCore? apsflyer, LemniskCleverTapCore? clevertap) async {
    WidgetsFlutterBinding.ensureInitialized();
    //lemnisk server...
    _writeKey = writeKey;
    _serverUrl = serverUrl;

    //appsflyer
    _enableAppsFlyer = apsflyer!.enable;
    _afDevKey = apsflyer.afDevKey;
    _afAppId = apsflyer.afAppId;
    if (_enableAppsFlyer == true) {
      try {
        Map appsFlyerOptions = {
          "afDevKey": _afDevKey,
          "afAppId": _afAppId,
          "isDebug": true,
        };
        _appsflyerSdk = AppsflyerSdk(appsFlyerOptions);
        _appsflyerSdk?.initSdk(
          registerConversionDataCallback: true,
          registerOnAppOpenAttributionCallback: true,
          registerOnDeepLinkingCallback: true,
        );
      } on Exception catch (e) {
        print("Lemnisk AppsFlyer initialization Error: ${e.toString()}");
      }
      print('Lemnisk AppsFlyer Enabled');
    } else {
      print('Lemnisk AppsFlyer Disabled');
    }

    //firebase core....
    _enableFirebase = firebase!.enable;
    _projectName = firebase.projectName;
    _appId = firebase.appId;
    _apiKey = firebase.apiKey;
    _projectId = firebase.projectId;
    _messagingSenderId = firebase.messagingSenderId;
    if (_enableFirebase == true) {
      try {
        FirebaseOptions firebaseOptions = FirebaseOptions(
          appId: _appId!,
          apiKey: _apiKey!,
          projectId: _projectId!,
          messagingSenderId: _messagingSenderId!
        );
        _appFirebase = await Firebase.initializeApp(name: _projectName!, options: firebaseOptions);
        analytics = FirebaseAnalytics.instance;
        // analytics?.app.setAutomaticDataCollectionEnabled(false);
      } on Exception catch (e) {
        print("Lemnisk Firebase intialization Error: ${e.toString()}");
      }
      print('Initialization Lemnisk Firebase Enabled');
    } else {
      print('Lemnisk Firebase Disabled');
    }

    _enableClevertap = clevertap!.enable;
    if(_enableClevertap == true) {
      try {
        CleverTapPlugin.activate();
        if (clevertap.isDebug == true) {
          CleverTapPlugin.setDebugLevel(1);
        }
        print('Lemnisk Clevertap Enabled');
      } on Exception catch (e) {
        print("Lemnisk Clevertap intialization Error: ${e.toString()}");
      }
    } else {
      print('Lemnisk Clevertap Disabled');
    }

    await getTransformConfig();

   /* await Workmanager().initialize(
      callbackDispatcher,
      isInDebugMode: true,
    );
    await Workmanager().registerPeriodicTask(
      "5",
      simplePeriodic30minsTask,
      frequency: Duration(minutes: 15),
    );*/
  }

  static Future<void> getTransformConfig() async {
    try {
      String url = _transformUrl + _writeKey! + "?a1=flutter&a2=" + sdkVersion;
      print('Lemnisk transform url: $url');
      Uri uri = Uri.parse(url);
      var response = await http.get(uri);
      if (response.statusCode == 200) {
        final String responseString = response.body;
        parsedWriteKeytoConfigMap = jsonDecode(responseString);
        print('Lemnisk transform config get success =====   $responseString');
      } else {
        final String responseString = response.body;
        print('Lemnisk transform config get failed =====   $responseString');
        parsedWriteKeytoConfigMap = jsonDecode(writeKeytoConfigMapping);
      }
    } on Exception catch (e) {
      print("Lemnisk transform config Error: ${e.toString()}");
      parsedWriteKeytoConfigMap = jsonDecode(writeKeytoConfigMapping);
    }
  }

  static final DeviceInfoPlugin _deviceInfoPlugin = DeviceInfoPlugin();
  //get Device ID
  static Future<String?> get _getDeviceId async {
    String? deviceId;
    try {
      if (Platform.isAndroid) {
        AndroidDeviceInfo androidInfo = await _deviceInfoPlugin.androidInfo;
        deviceId = androidInfo.androidId;
      } else if (Platform.isIOS) {
        IosDeviceInfo iosInfo = await _deviceInfoPlugin.iosInfo;
        deviceId = iosInfo.identifierForVendor;
      } else {
        deviceId = (await PlatformDeviceIdPlatform.instance.getDeviceId());
      }
    } on PlatformException {
      deviceId = '';
    }
    return deviceId;
  }

  //get Plateform Device Type
  static Future<String?> get _getPlatFormDeviceType async {
    String? deviceType;
    try {
      if (Platform.isAndroid) {
        deviceType = "Android";
      } else if (Platform.isIOS) {
        deviceType = "IOS";
      }
    } on PlatformException {
      deviceType = '';
    }
    return deviceType;
  }

  //get Device Model Name
  static Future<String?> get _getDeviceModel async {
    String? deviceModel = '';
    try {
      if (Platform.isAndroid) {
        AndroidDeviceInfo androidInfo = await _deviceInfoPlugin.androidInfo;
        deviceModel = androidInfo.model;
      } else if (Platform.isIOS) {
        IosDeviceInfo iosInfo = await _deviceInfoPlugin.iosInfo;
        deviceModel = iosInfo.name;
      } else {
        deviceModel = (await PlatformDeviceIdPlatform.instance.getDeviceId());
      }
    } on PlatformException {
      deviceModel = '';
    }
    return deviceModel;
  }

  //get Device Manufacture Model
  static Future<String?> get _getManufactureModel async {
    String? deviceManufacture = '';
    try {
      if (Platform.isAndroid) {
        AndroidDeviceInfo androidInfo = await _deviceInfoPlugin.androidInfo;

        deviceManufacture = androidInfo.manufacturer;
      } else if (Platform.isIOS) {
        IosDeviceInfo iosInfo = await _deviceInfoPlugin.iosInfo;
        deviceManufacture = iosInfo.systemName;
      } else {
        deviceManufacture = (await PlatformDeviceIdPlatform.instance.getDeviceId());
      }
    } on PlatformException {
      deviceManufacture = '';
    }
    return deviceManufacture;
  }

  //get Device OS Version
  static Future<String?> get _getOSVersion async {
    String? deviceID = '';
    try {
      if (Platform.isAndroid) {
        AndroidDeviceInfo androidInfo = await _deviceInfoPlugin.androidInfo;
        deviceID = androidInfo.version.release;
      } else if (Platform.isIOS) {
        IosDeviceInfo iosInfo = await _deviceInfoPlugin.iosInfo;
        deviceID = iosInfo.systemVersion;
      } else {
        deviceID = (await PlatformDeviceIdPlatform.instance.getDeviceId());
      }
    } on PlatformException {
      deviceID = '';
    }
    return deviceID;
  }

  //get Device Google Advertising ID
  static Future<String?> get _getGoogleadvertisingID async {
    String? advertisingId;
    try {
      if (Platform.isAndroid) {
        try {
          advertisingId = await AdvertisingId.id(true);
        } on PlatformException {
          advertisingId = 'Failed to get platform version.';
        }
      } else if (Platform.isIOS) {
        advertisingId = await AdvertisingId.id(true);
      } else {
        advertisingId = (await PlatformDeviceIdPlatform.instance.getDeviceId());
      }
    } on PlatformException {
      advertisingId = '';
    }
    return advertisingId;
  }

  //get Device Google Advertising bool Value
  static Future<bool?> get _getGoogleadvertisingIDBoolValue async {
    bool? isLimitAdTrackingEnabled;
    try {
      if (Platform.isAndroid) {
        try {
          isLimitAdTrackingEnabled = await AdvertisingId.isLimitAdTrackingEnabled;
        } on PlatformException {
          isLimitAdTrackingEnabled = false;
        }
      } else if (Platform.isIOS) {
        isLimitAdTrackingEnabled = await AdvertisingId.isLimitAdTrackingEnabled;
      }
    } on PlatformException {
      isLimitAdTrackingEnabled = false;
    }
    return isLimitAdTrackingEnabled;
  }

  //get Device Width
  static Future<String?> get _getScreenWidth async {
    var physicalScreenSize = window.physicalSize;
    var physicalWidth = physicalScreenSize.width;
    return physicalWidth.toString();
  }

  //get Device Height
  static Future<String> get _getScreenHeight async {
    var physicalScreenSize = window.physicalSize;
    var physicalHeight = physicalScreenSize.height;
    return physicalHeight.toString();
  }

  //get App Name
  static Future<String> get _getAppName async {
    PackageInfo _packageInfo = PackageInfo(
      appName: 'Unknown',
      packageName: 'Unknown',
      version: 'Unknown',
      buildNumber: 'Unknown',
      buildSignature: 'Unknown',
    );
    final info = await PackageInfo.fromPlatform();
    _packageInfo = info;
    return _packageInfo.appName;
  }

  //get App Version
  static Future<String> get _getAppVersion async {
    PackageInfo _packageInfo = PackageInfo(
      appName: 'Unknown',
      packageName: 'Unknown',
      version: 'Unknown',
      buildNumber: 'Unknown',
      buildSignature: 'Unknown',
    );
    final info = await PackageInfo.fromPlatform();
    _packageInfo = info;
    return _packageInfo.version;
  }

  //get App Build No
  static Future<String> get _getBuildNumber async {
    PackageInfo _packageInfo = PackageInfo(
      appName: 'Unknown',
      packageName: 'Unknown',
      version: 'Unknown',
      buildNumber: 'Unknown',
      buildSignature: 'Unknown',
    );
    final info = await PackageInfo.fromPlatform();
    _packageInfo = info;
    return _packageInfo.buildNumber;
  }

  //get Device Network IP
  static Future<String?> get _getNetworkIP async {
    final url = Uri.parse('https://api.ipify.org/');
    final response = await http.get(url);
    return response.body;
  }

  //get Payload Received Time
  static Future<dynamic> get _getReceivedAt async {
    return DateTime.now().toUtc().toString();
  }

  //get Device Payload Send Time
  static Future<String?> get _getSendAt async {
    DateTime getCurrentTimestamp = DateTime.now();
    String agreementSigned = formatISOTime(getCurrentTimestamp);
    return agreementSigned;
  }

  //get Device TimesTemp
  static Future<String?> get _getTimestemp async {
    DateTime getCurrentTimestamp = DateTime.now();
    String agreementSigned = formatISOTime(getCurrentTimestamp);
    return agreementSigned;
  }

  static String formatISOTime(DateTime date) {
    var duration = date.timeZoneOffset;
    if (duration.isNegative)
      return (DateFormat("yyyy-MM-ddTHH:mm:ss.mmm").format(date) + "-${duration.inHours.toString().padLeft(2, '0')}${(duration.inMinutes - (duration.inHours * 60)).toString().padLeft(2, '0')}");
    else
      return (DateFormat("yyyy-MM-ddTHH:mm:ss.mmm").format(date) + "+${duration.inHours.toString().padLeft(2, '0')}${(duration.inMinutes - (duration.inHours * 60)).toString().padLeft(2, '0')}");
  }

  //get Device Original TimesTemp
  static Future<String?> get _getOriginalTimestemp async {
    DateTime getCurrentTimestamp = DateTime.now();
    String agreementSigned = formatISOTime(getCurrentTimestamp);
    return agreementSigned;
  }

  //get Device Original TimesTemp
  static Future<String?> get _getServerTs async {
    return DateTime.now().toUtc().toString();
  }

  //get Device Original TimesTemp
  static Future<String?> get _getMessageId async {
    var uuid = Uuid();
    return uuid.v4();
  }

  static void track(String name, AttributeBuilder properties, OtherBuilder otherid) async {
    GenricModel genricModel = GenricModel();
    genricModel.name = name;
    genricModel.properties = properties.properties;
    genricModel.otherid = otherid.otherid;
    GenricModel? fMain = await genricModel;

    String? deviceId;
    String? deviceType;
    String? deviceManufacture;
    String? deviceModel;
    String? versionName;

    String? appName;
    String? appVersion;
    String? appBuild;

    String? googleAdvertisingID;
    dynamic? googleAdvertisingBool;

    String? networkIP;

    String? appScreenHeightWidth;
    String? appScreenHeightHeight;

    String? appDataReceivedAt;
    String? appDataSenddAt;
    String? appOriginalTimetemp;
    String? appTimestemp;
    String? appServerTs;
    String? appMsgId;

    try {
      deviceId = await _getDeviceId;
      deviceType = await _getPlatFormDeviceType;
      deviceModel = await _getDeviceModel;
      deviceManufacture = await _getManufactureModel;
      versionName = await _getOSVersion;
      appName = await _getAppName;
      appVersion = await _getAppVersion;
      appBuild = await _getBuildNumber;
      googleAdvertisingID = await _getGoogleadvertisingID;
      googleAdvertisingBool = await _getGoogleadvertisingIDBoolValue;
      networkIP = await _getNetworkIP;

      appScreenHeightHeight = await _getScreenHeight;
      appScreenHeightWidth = await _getScreenWidth;

      appDataReceivedAt = await _getReceivedAt;
      appDataSenddAt = await _getSendAt;

      appOriginalTimetemp = await _getOriginalTimestemp;
      appServerTs = await _getServerTs;
      appTimestemp = await _getTimestemp;
      appMsgId = await _getMessageId;
    } on PlatformException {
      deviceId = 'Failed to get information.';
    }

    Map<dynamic, dynamic> lib = {
      "library": {
        "name": "Flutter SDK",
        "version": sdkVersion,
      },
      "server_ts": appServerTs,
      "app": {
        "name": appName,
        "version": appVersion,
        "build": appBuild,
      },
      "device": {
        "id": deviceId,
        "advertisingId": googleAdvertisingID,
        "adTrackingEnabled": googleAdvertisingBool,
        "manufacturer": deviceManufacture,
        "model": deviceModel,
        "type": deviceType,
        "token": LemniskFirebase.FCM_URL,
      },
      "screen": {
        "width": appScreenHeightWidth,
        "height": appScreenHeightHeight,
        "density": 2,
      },
      "userAgent": {
        "osType": deviceType,
        "osVersion": versionName,
      },
      "ip": networkIP
    };
    Map map = {
      "id": deviceId,
      "messageId": appMsgId,
      "userId": "userId",
      "otherIds": fMain.otherid,
      "event": name,
      "properties": fMain.properties,
      "receivedAt": appDataReceivedAt,
      "sentAt": appDataSenddAt,
      "timestamp": appTimestemp,
      "type": "track",
      "originalTimestamp": appOriginalTimetemp,
      "writeKey": _writeKey,
    };
    map.addAll({"context": lib});

    //After the full payload is created..now we send all to lemnisk server
    print('track event' + _serverUrl!);
    Uri uri = Uri.parse(_serverUrl!);
    var response = await http.post(uri, body: JsonEncoder().convert(map));
    if (response.statusCode == 200) {
      final String responseString = response.body;
      print('event track success =====   $responseString');
    } else {
      final String responseString = response.body;
      print('event track failed =====   $responseString');
    }

    //appsflyer analytics
    if (_enableAppsFlyer == true) {
      try {
        trackAppsFlyerEvent(name, properties.properties);
      } on Exception catch (e) {
        print("logEvent Appsflyer Error: ${e.toString()}");
      }
    }

    //clevertap analytics
    if (_enableClevertap == true) {
      try {
        trackClevertapEvent(name, properties.properties);
      } on Exception catch (e) {
        print("recordEvent Clevertap Error: ${e.toString()}");
      }
    }

    //firebase analytics
    if (_enableFirebase == true) {
      try {
        trackGaEvent(name, properties.properties);
        print("logEvent GA4 success");
      } on Exception catch (e) {
        print("logEvent GA4 Error: ${e.toString()}");
      }
    }
  }

  static void identify(String userID, AttributeBuilder properties, OtherBuilder otherid) async {
    GenricModel genricModel = GenricModel();
    genricModel.name = userID;
    genricModel.properties = properties.properties;
    genricModel.otherid = otherid.otherid;
    GenricModel? fMain = await genricModel;

    String? deviceId;
    String? deviceType;
    String? deviceManufacture;
    String? deviceModel;
    String? versionName;

    String? appName;
    String? appVersion;
    String? appBuild;

    String? googleAdvertisingID;
    dynamic? googleAdvertisingBool;

    String? networkIP;

    String? appScreenHeightWidth;
    String? appScreenHeightHeight;

    String? appDataReceivedAt;
    String? appDataSenddAt;
    String? appOriginalTimetemp;
    String? appTimestemp;
    String? appServerTs;
    String? appMsgId;

    try {
      deviceId = await _getDeviceId;
      deviceType = await _getPlatFormDeviceType;
      deviceModel = await _getDeviceModel;
      deviceManufacture = await _getManufactureModel;
      versionName = await _getOSVersion;
      appName = await _getAppName;
      appVersion = await _getAppVersion;
      appBuild = await _getBuildNumber;
      googleAdvertisingID = await _getGoogleadvertisingID;
      googleAdvertisingBool = await _getGoogleadvertisingIDBoolValue;
      networkIP = await _getNetworkIP;

      appScreenHeightHeight = await _getScreenHeight;
      appScreenHeightWidth = await _getScreenWidth;

      appDataReceivedAt = await _getReceivedAt;
      appDataSenddAt = await _getSendAt;

      appOriginalTimetemp = await _getOriginalTimestemp;
      appTimestemp = await _getTimestemp;
      appServerTs = await _getServerTs;
      appMsgId = await _getMessageId;
    } on PlatformException {
      deviceId = 'Failed to get information.';
    }

    Map<dynamic, dynamic> lib = {
      "library": {
        "name": "Flutter SDK",
        "version": sdkVersion,
      },
      "server_ts": appServerTs,
      "app": {
        "name": appName,
        "version": appVersion,
        "build": appBuild,
      },
      "device": {
        "id": deviceId,
        "advertisingId": googleAdvertisingID,
        "adTrackingEnabled": googleAdvertisingBool,
        "manufacturer": deviceManufacture,
        "model": deviceModel,
        "type": deviceType,
        "token": LemniskFirebase.FCM_URL,
      },
      "screen": {
        "width": appScreenHeightWidth,
        "height": appScreenHeightHeight,
        "density": 2,
      },
      "userAgent": {
        "osType": deviceType,
        "osVersion": versionName,
      },
      "ip": networkIP
    };
    Map map = {
      "id": deviceId,
      "userId": "userId",
      "otherIds": fMain.otherid,
      "event": userID,
      "messageId": appMsgId,
      "customerProperties": fMain.properties,
      "receivedAt": appDataReceivedAt,
      "sentAt": appDataSenddAt,
      "timestamp": appTimestemp,
      "type": "identify",
      "originalTimestamp": appOriginalTimetemp,
      "writeKey": _writeKey,
    };
    map.addAll({"context": lib});

    //After the full payload is created..now we send all to lemnisk server
    print('identify event' + _serverUrl!);
    Uri uri = Uri.parse(_serverUrl!);
    var response = await http.post(uri, body: JsonEncoder().convert(map));
    if (response.statusCode == 200) {
      final String responseString = response.body;
      print('event identify success =====   $responseString');
    } else {
      final String responseString = response.body;
      print('event identify failed =====   $responseString');
    }

    if (_enableAppsFlyer == true) {
      try {
        _appsflyerSdk?.setCustomerUserId(userID);
        print('AppsFlyer userId set as $userID');
        // _appsflyerSdk?.setUserEmails(emails)
      } on Exception catch (e) {
        print("Result logEvent appsflyer Error: ${e.toString()}");
      }
    }

    //clevertap analytics
    if (_enableClevertap == true) {
      try {
        Map<String, dynamic> ctMap = new Map.from(properties.properties);
        ctMap.putIfAbsent("Identity", () => userID);
        if(otherid.otherid["email"] != null) {
          ctMap.putIfAbsent("Email", () => otherid.otherid["email"]);
          otherid.otherid.remove("email");
        }
        if(otherid.otherid["mobile"] != null) {
          ctMap.putIfAbsent("Phone", () => otherid.otherid["mobile"]);
          otherid.otherid.remove("mobile");
        }
        if(otherid.otherid["phone"] != null) {
          ctMap.putIfAbsent("Phone", () => otherid.otherid["phone"]);
          otherid.otherid.remove("phone");
        }
        ctMap.addAll(otherid.otherid);
        CleverTapPlugin.profileSet(ctMap);
      } on Exception catch (e) {
        print("recordEvent CleverTap Error: ${e.toString()}");
      }
    }

    if (_enableFirebase == true) {
      //firebase analytics
      analytics?.setUserId(id: userID);
      print("setUserId firebase");
    }
  }

  static void screen(String screenName, AttributeBuilder properties, OtherBuilder otherid) async {
    GenricModel genricModel = GenricModel();
    genricModel.name = screenName;
    genricModel.properties = properties.properties;
    genricModel.otherid = otherid.otherid;
    GenricModel? fMain = await genricModel;

    String? deviceId;
    String? deviceType;
    String? deviceManufacture;
    String? deviceModel;
    String? versionName;

    String? appName;
    String? appVersion;
    String? appBuild;

    String? googleAdvertisingID;
    dynamic? googleAdvertisingBool;

    String? networkIP;

    String? appScreenHeightWidth;
    String? appScreenHeightHeight;

    String? appDataReceivedAt;
    String? appDataSenddAt;
    String? appOriginalTimetemp;
    String? appTimestemp;
    String? appServerTs;
    String? appMsgId;

    try {
      deviceId = await _getDeviceId;
      deviceType = await _getPlatFormDeviceType;
      deviceModel = await _getDeviceModel;
      deviceManufacture = await _getManufactureModel;
      versionName = await _getOSVersion;
      appName = await _getAppName;
      appVersion = await _getAppVersion;
      appBuild = await _getBuildNumber;
      googleAdvertisingID = await _getGoogleadvertisingID;
      googleAdvertisingBool = await _getGoogleadvertisingIDBoolValue;
      networkIP = await _getNetworkIP;

      appScreenHeightHeight = await _getScreenHeight;
      appScreenHeightWidth = await _getScreenWidth;

      appDataReceivedAt = await _getReceivedAt;
      appDataSenddAt = await _getSendAt;

      appOriginalTimetemp = await _getOriginalTimestemp;
      appTimestemp = await _getTimestemp;
      appServerTs = await _getServerTs;
      appMsgId = await _getMessageId;
    } on PlatformException {
      deviceId = 'Failed to get information.';
    }

    Map<dynamic, dynamic> lib = {
      "library": {
        "name": "Flutter SDK",
        "version": sdkVersion,
      },
      "server_ts": appServerTs,
      "app": {
        "name": appName,
        "version": appVersion,
        "build": appBuild,
      },
      "device": {
        "id": deviceId,
        "advertisingId": googleAdvertisingID,
        "adTrackingEnabled": googleAdvertisingBool,
        "manufacturer": deviceManufacture,
        "model": deviceModel,
        "type": deviceType,
        "token": LemniskFirebase.FCM_URL,
      },
      "screen": {
        "width": appScreenHeightWidth,
        "height": appScreenHeightHeight,
        "density": 2,
      },
      "userAgent": {
        "osType": deviceType,
        "osVersion": versionName,
      },
      "ip": networkIP
    };

    Map map = {
      "id": deviceId,
      "userId": "userId",
      "otherIds": fMain.otherid,
      "event": screenName,
      "messageId": appMsgId,
      "properties": fMain.properties,
      "receivedAt": appDataReceivedAt,
      "sentAt": appDataSenddAt,
      "timestamp": appTimestemp,
      "type": "screen",
      "originalTimestamp": appOriginalTimetemp,
      "writeKey": _writeKey,
    };
    map.addAll({"context": lib});

    //After the full payload is created..now we send all to lemnisk server
    print('Screen event' + _serverUrl!);
    Uri uri = Uri.parse(_serverUrl!);
    var response = await http.post(uri, body: JsonEncoder().convert(map));
    if (response.statusCode == 200) {
      final String responseString = response.body;
      print('event Screen success =====   $responseString');
    } else {
      final String responseString = response.body;
      print('event Screen failed =====   $responseString');
    }

    //appsflyer analytics
    if (_enableAppsFlyer == true) {
      bool? result;
      try {
        result = await _appsflyerSdk?.logEvent("screenView: $screenName", properties.properties);
        if (result == true) {
          print("logScreenView appsflyer success: ${result}");
        } else {
          print("logScreenView appsflyer failed: ${result}");
        }
      } on Exception catch (e) {
        print("logScreenView appsflyer Error: ${e.toString()}");
      }
    }

    //clevertap analytics
    if (_enableClevertap == true) {
      try {
        trackClevertapEvent("screenView: $screenName", properties.properties);
      } on Exception catch (e) {
        print("recordEvent CleverTap Error: ${e.toString()}");
      }
    }

    if (_enableFirebase == true) {
      //firebase analytics
      properties.properties.putIfAbsent("screen_name", () => screenName);
      analytics?.logEvent(name: "screen_view", parameters: properties.properties);
      print("GA logScreenView: " + screenName);
    }
  }

  static Map<String, dynamic> recurseAndPopulate(
      Map<String, dynamic> gaProp, Map<String, dynamic>? properties) {
    Iterable<String> keys = gaProp.keys;
    Map<String, dynamic> returnValue = {};
    for (String index in keys) {
      dynamic value = gaProp[index];
      dynamic updatedValue;
      // Type of value can be String or List or Map
      if (value is String) {
        updatedValue = getStringValue(value, properties);
        if(properties != null) {
          properties[index] = updatedValue;
        }
      } else if (value is Map) {
        updatedValue =
            recurseAndPopulate(value as Map<String, dynamic>, properties);
        if(properties != null) {
          properties[index] = updatedValue;
        }
      } else if (value is List) {
        updatedValue = getListValue(value, properties);
        if(properties != null) {
          properties[index] = updatedValue;
        }
      }
      returnValue[index] = updatedValue;
    }
    return returnValue;
  }

  static dynamic getListValue(List<dynamic> value, Map<String, dynamic>? properties) {
    List<dynamic> returnValue = [];
    for (dynamic element in value) {
      dynamic updatedValue = element;
      if (element is String) {
        updatedValue = getStringValue(element, properties);
      } else if (element is Map) {
        updatedValue = recurseAndPopulate(element as Map<String, dynamic>, properties);
      } else if (element is List) {
        updatedValue = getListValue(element, properties);
      }
      returnValue.add(updatedValue);
    }
    return returnValue;
  }

  static dynamic getStringValue(String value, Map<String, dynamic>? properties) {
    if (value.startsWith('properties')) {
      List<String> fields = value.split('.');
      return getValue(properties, fields.sublist(1, fields.length));
    }
    return value;
  }

  static dynamic getValue(Map<String, dynamic>? properties, List<String>? fields) {
    if (properties != null && fields != null) {
      if (fields.length == 1) {
        if (properties.containsKey(fields[0])) {
          return properties[fields[0]];
        } else {
          return "";
        }
      }
      bool hasOwnprop = properties.containsKey(fields[0]);
      bool typo = properties[fields[0]] is Map;
      if (hasOwnprop && typo) {
        return getValue(properties[fields[0]], fields.sublist(1, fields.length));
      }
    }
    return "";
  }

  static void trackGaEvent(String eventName, Map<String, dynamic> properties) {
    String gaName = eventName;
    dynamic gaProp = parsedWriteKeytoConfigMap!["ga4"][eventName];
    Map<String, dynamic> val = properties;

    if (gaProp == null) {
      print("No mapping found for GA4, send as it is");
    } else if (gaProp is String) {
      print("GA4 eventname String conversion");
      gaName = gaProp;
    } else if (gaProp is Map) {
      print("GA4 Mapping found");
      Map<String, dynamic> gaProp1 = gaProp as Map<String, dynamic>;
      gaName = gaProp1.keys.first;
      val = recurseAndPopulate(gaProp1[gaName], properties);
    } else if (gaProp is List) {
      print("gaProp is List");
    } else {
      print("Unknown mapping to GA4 event");
    }
    print("GA: " + gaName + ", properties: " + val.toString());
    analytics?.logEvent(name: gaName, parameters: val);
  }

  static void trackAppsFlyerEvent(String eventName, Map<String, dynamic> properties) {
    String apfName = parsedWriteKeytoConfigMap!["appsflyer"]["event"][eventName] ?? eventName;
    Iterable<String> keys = properties.keys;
    bool? result;
    Map<String, dynamic> returnValue = {};
    for (String index in keys) {
      String? returnKey = parsedWriteKeytoConfigMap!["appsflyer"]["properties"][index];
      if(returnKey != null) {
        returnValue[returnKey] = properties[index];
      } else {
        returnValue[index] = properties[index];
      }
    }
    print("apfName: " + apfName + ", properties: " + returnValue.toString());
    _appsflyerSdk?.logEvent(apfName, returnValue);
  }

  static void trackClevertapEvent(String eventName, Map<String, dynamic> properties) {
    print("ctName: " + eventName + ", properties: " + properties.toString());
    List<dynamic> ctChargeEvents = parsedWriteKeytoConfigMap!["clevertap"]["chargeEvents"];
    if(ctChargeEvents.contains(eventName)) {
      List<dynamic> ctChargeDetailsParams = parsedWriteKeytoConfigMap!["clevertap"]["chargeDetails"];
      Map<String, dynamic> chargeDetails = {};
      Map<String, dynamic> item = {};
      Iterable<String> keys = properties.keys;
      for (String key in keys) {
        if(ctChargeDetailsParams.contains(key)) {
          chargeDetails[key] = properties[key];
        } else {
          item[key] = properties[key];
        }
      }
      CleverTapPlugin.recordChargedEvent(chargeDetails, [item]);
    } else {
      CleverTapPlugin.recordEvent(eventName, properties);
    }
  }
}
